/* always defined to indicate that i18n is enabled */
#mesondefine ENABLE_NLS

/* The gettext translation domain */
#mesondefine GETTEXT_PACKAGE

/* Define if the xrandr-$XRANDR_REQUIRED library is present */
#mesondefine HAVE_RANDR

/* Building with startup notification support */
#mesondefine HAVE_STARTUP_NOTIFICATION

/* Enable additional debugging at the expense of performance and size */
#mesondefine MATE_ENABLE_DEBUG

/* Define to the major version */
#mesondefine MATE_MAJOR

/* Define to the minor version */
#mesondefine MATE_MINOR

/* Define to the micro version */
#mesondefine MATE_MICRO

/* Name of package */
#mesondefine PACKAGE

/* Define to the address where bug reports for this package should be sent. */
#mesondefine PACKAGE_BUGREPORT

/* Define to the full name of this package. */
#mesondefine PACKAGE_NAME

/* Define to the full name and version of this package. */
#mesondefine PACKAGE_STRING

/* Define to the one symbol short name of this package. */
#mesondefine PACKAGE_TARNAME

/* Define to the home page for this package. */
#mesondefine PACKAGE_URL

/* Define to the version of this package. */
#mesondefine PACKAGE_VERSION

/* Version number of package */
#mesondefine VERSION

/* Define to 1 if the X Window System is missing or not being used. */
#mesondefine X_DISPLAY_MISSING
